import datetime
import json
import math
import re
import subprocess as sp
import time
from binascii import unhexlify
from multiprocessing import Queue
from typing import List

import serial
from loguru import logger
from typing_extensions import Literal

from homesolar.utils import bluetooth

BYTE_ORDER = Literal["little", "big"]


class Parameter:
    def __init__(self, field: str, start_byte: int, stop_byte: int, byte_order: BYTE_ORDER = "big",
                 signed: bool = False, precision: float = 1.0):
        self.field = field
        self.start_byte = start_byte
        self.stop_byte = stop_byte
        self.byte_order = byte_order
        self.signed = signed
        self.precision = precision

    def get_value(self, reading_data):
        data = reading_data[self.start_byte: self.stop_byte]
        value = int.from_bytes(data[self.start_byte: self.stop_byte], byteorder=self.byte_order,
                               signed=self.signed) * self.precision
        if self.precision < 1:
            value = round(value, math.ceil(math.log(1 / self.precision, 10)))

        return value

    def get_bits(self, reading_data):
        string_bytes = reading_data[self.start_byte: self.stop_byte]

        bytes_list = []
        for i in range(len(string_bytes)):
            bytes_list.append(string_bytes[i:i + 1])

        bits = []
        for byte in bytes_list:
            for i in range(8):
                bits.append((ord(byte) >> i) & 1)
        return bits


class BluetoothClient:
    def __init__(self, address: str, baudrate: int = 9600, parity: str = serial.PARITY_NONE,
                 stopbits: int = serial.STOPBITS_ONE, bytesize: int = serial.EIGHTBITS, params: List[Parameter] = None,
                 request_code: str = "", read_size: int = 0, read_period: int = 5):

        self.address = address
        self.baudrate = baudrate
        self.parity = parity
        self.stopbits = stopbits
        self.bytesize = bytesize
        self.params = params
        self.request_code = request_code
        self.read_size = read_size
        self.read_period = read_period

        self._verification = None
        self._calculated_data = None

        self.connection = serial.Serial()
        self.connect()

    @property
    def verification(self):
        return self._verification

    @verification.setter
    def verification(self, func):
        self._verification = func

    @property
    def calculated_data(self):
        return self._calculated_data

    @calculated_data.setter
    def calculated_data(self, func):
        self._calculated_data = func

    def set_params(self, params: list):
        self.params = params

    def set_request_code(self, request_code: str):
        self.request_code = request_code

    def set_read_size(self, read_size: int):
        self.read_size = read_size

    def set_read_period(self, read_period: int):
        self.read_period = read_period

    def set_verification(self, verification: dict):
        self.verification = verification

    def connect(self):
        if not self.is_connected():
            while self.rfcomm_port() == -1:
                if not self.bind_rfcomm():
                    logger.warning(
                        "Cannot bind Bluetooth Device to a RFComm Address. Please make sure that the device is already paired")
                    break

            self.connection = serial.Serial(
                port=f"dev/rfcomm{self.rfcomm_port()}",
                baudrate=self.baudrate,
                parity=self.parity,
                stopbits=self.stopbits,
                bytesize=self.bytesize)

            if not self.connection.isOpen():
                self.connection.open()

    def disconnect(self):
        if self.connection.isOpen():
            self.connection.close()

        if not self.release_rfcomm():
            logger.warning(
                "Cannot release Bluetooth Device from the RFComm Address. Might want to manually verify")

    def rfcomm_port(self):
        stdoutdata = sp.getoutput("sudo rfcomm")
        pattern = re.compile(r"rfcomm(\d+): (\S+)")

        # Split the output into lines and search for the target MAC address
        for line in stdoutdata.split('\n'):
            match = pattern.search(line)
            if match:
                port, mac_address = match.groups()
                if mac_address == self.address:
                    return port

        return -1

    def bind_rfcomm(self):
        for port_number in range(0, 30):
            stdoutdata = sp.getoutput(f"sudo rfcomm")
            if self.address in stdoutdata:
                return True
            stdoutdata = sp.getoutput(f"sudo rfcomm bind dev/rfcomm{port_number} {self.address}")
            if "Can't create device" not in stdoutdata.split():
                return True
        return False

    def release_rfcomm(self):
        if self.rfcomm_port() != -1:
            stdoutdata = sp.getoutput(f"sudo rfcomm release dev/rfcomm{self.rfcomm_port()}")

            if "Can't release device" not in stdoutdata.split():
                return True
        return False

    def is_connected(self):
        stdoutdata = sp.getoutput("hcitool con")

        if self.address in stdoutdata.split():
            return True
        return False

    def run(self, task_queue: Queue):
        try:
            if self.read_size == 0 or self.params is None:
                raise Exception("Bluetooth read_size is 0, please make sure to setup the device correctly")

            while True:
                if not self.is_connected():
                    self.connect()

                if not self.connection.isOpen():
                    self.connection.open()
                else:
                    try:
                        if self.request_code != "":
                            self.connection.write(unhexlify(self.request_code))

                        read_data = self.connection.read(self.read_size)

                        fields = {}
                        verified = True
                        for param in self.params:
                            fields[param.field] = param.get_value(read_data)

                            if self.calculated_data is not None:
                                self.calculated_data(param, read_data, fields)

                            if self.verification is not None:
                                if not self.verification(param, read_data):
                                    verified = False
                                    self.connection.close()
                                    time.sleep(1)
                                    break

                        if verified:
                            data = {
                                "name": "Antw33-BMS",
                                "payload": json.dumps(fields),
                                'time': datetime.datetime.now().timestamp()
                            }
                            logger.debug(f"Incoming Sensor Data [{data['name']}]")
                            task = {
                                "name": "write_sensor_data",
                                "data": data
                            }
                            task_queue.put(task)
                            time.sleep(5)

                    except Exception as e:
                        logger.warning(f"Something failing during bluetooth communication, [{e}]")

        except Exception as e:
            logger.exception(f"Something went wrong when dealing with bluetooth connection, [{e}]")
        finally:
            self.disconnect()


def initialize(main_task_queue: Queue):
    params = [
        Parameter("Voltage", 4, 6, signed=True, precision=0.1),
        Parameter("Current", 70, 74, signed=True, precision=0.1),
        Parameter("Power", 111, 115, signed=True),
        Parameter("Capacity", 75, 79, signed=False, precision=0.000001),
        Parameter("ChargeStatus", 103, 104),
        Parameter("DischargeStatus", 104, 105),
        Parameter("BalanceStatus", 105, 106),
        Parameter("BalanceCells", 133, 136),
        Parameter("Temp1", 91, 93),
        Parameter("Temp2", 93, 95),
        Parameter("Temp3", 95, 97),
        Parameter("Temp4", 97, 99),
        Parameter("Temp5", 99, 101),
        Parameter("Temp6", 101, 103),
        Parameter("MinCell", 118, 119),
        Parameter("MaxCell", 115, 116),
        Parameter("MinVolt", 119, 121, signed=False, precision=0.001),
        Parameter("MaxVolt", 116, 118, signed=False, precision=0.001),
        Parameter("AvgVolt", 121, 123, precision=0.001),
        Parameter("Cell1", 6, 8, signed=False, precision=0.001),
        Parameter("Cell2", 8, 10, signed=False, precision=0.001),
        Parameter("Cell3", 10, 12, signed=False, precision=0.001),
        Parameter("Cell4", 12, 14, signed=False, precision=0.001),
        Parameter("Cell5", 14, 16, signed=False, precision=0.001),
        Parameter("Cell6", 16, 18, signed=False, precision=0.001),
        Parameter("Cell7", 18, 20, signed=False, precision=0.001),
        Parameter("Cell8", 20, 22, signed=False, precision=0.001),
        Parameter("Cell9", 22, 24, signed=False, precision=0.001),
        Parameter("Cell10", 24, 26, signed=False, precision=0.001),
        Parameter("Cell11", 26, 28, signed=False, precision=0.001),
        Parameter("Cell12", 28, 30, signed=False, precision=0.001),
        Parameter("Cell13", 30, 32, signed=False, precision=0.001),
        Parameter("Cell14", 32, 34, signed=False, precision=0.001),
        Parameter("Cell15", 34, 36, signed=False, precision=0.001),
        Parameter("Cell16", 36, 38, signed=False, precision=0.001),
        Parameter("Cell17", 38, 40, signed=False, precision=0.001),
        Parameter("Cell18", 40, 42, signed=False, precision=0.001),
        Parameter("Cell19", 42, 44, signed=False, precision=0.001),
        Parameter("Cell20", 44, 46, signed=False, precision=0.001),
        Parameter("Cell21", 46, 48, signed=False, precision=0.001),
        Parameter("Cell22", 48, 50, signed=False, precision=0.001),
        Parameter("Cell23", 50, 52, signed=False, precision=0.001),
        Parameter("Cell24", 52, 54, signed=False, precision=0.001)
    ]

    ant_bms = BluetoothClient("AA:BB:CC:91:23:45")
    ant_bms.set_params(params)
    ant_bms.set_read_size(140)
    ant_bms.set_request_code("DBDB00000000")

    def verification(param: Parameter, reading_data):
        if param.field == "AvgVolt":
            if param.get_value(reading_data) == 0:
                return False
        return True

    def calculated_data(param: Parameter, reading_data, fields: dict):
        if param.field == 'BalanceCells':
            for index, bal in enumerate(param.get_bits(reading_data)):
                fields[f"Bal{index}"] = bal

        if param.field == 'AvgVolt':
            fields['SoC'] = bluetooth.ManualSoC().get_soc(param.get_value(reading_data))

    ant_bms.verification = verification
    ant_bms.calculated_data = calculated_data
    ant_bms.run(main_task_queue)
    # try:
    #     ser = serial.Serial(
    #         port=config.homesolar_config['BLUETOOTH']['port'],
    #         baudrate=9600,
    #         parity=serial.PARITY_NONE,
    #         stopbits=serial.STOPBITS_ONE,
    #         bytesize=serial.EIGHTBITS)
    #
    #     try:
    #         ser.open()
    #         time.sleep(1)
    #     except Exception as e:
    #         ser.close()
    #         ser.open()
    #         logger.warning(f"Something went wrong when trying to open Bluetooth as a Serial Connection [{e}]")
    #
    #     first_boot = True
    #     while True:
    #         try:
    #             ser.write(unhexlify(config.homesolar_config['BLUETOOTH']['request_code']))
    #             if first_boot:
    #                 time.sleep(1)
    #                 first_boot = False
    #
    #             antw33 = ser.read(140)
    #
    #             voltage = 0
    #             fields = {}
    #             for data in params:
    #                 fields[data.field] = data.value
    #
    #                 if data.field == 'AvgVolt':
    #                     voltage = data.value
    #
    #                 if data.field == 'BalanceCells':
    #                     for index, bal in enumerate(data.get_bits()):
    #                         fields[f"Bal{index}"] = bal
    #
    #                 if data.field == 'AvgVolt':
    #                     fields['SoC'] = bluetooth.ManualSoC().get_soc(data.value)
    #
    #             if voltage == 0:
    #                 ser.close()
    #                 time.sleep(1)
    #                 ser.open()
    #                 time.sleep(1)
    #
    #             data = {
    #                 "name": "Antw33-BMS",
    #                 "payload": json.dumps(fields),
    #                 'time': datetime.datetime.now().timestamp()
    #             }
    #             logger.debug(f"Incoming Sensor Data [{data['name']}]")
    #             task = {
    #                 "name": "write_sensor_data",
    #                 "data": data
    #             }
    #             main_task_queue.put(task)
    #             time.sleep(5)
    #         except:
    #             ser.close()
    #             time.sleep(1)
    #             ser.open()
    #             time.sleep(1)
    #
    # except Exception as ex:
    #     print(ex)
    # finally:
    #     try:
    #         ser.close()
    #     except Exception as e:
    #         logger.warning(f"Something went wrong when trying to close Bluetooth as Serial Connection [{e}]")



from loguru import logger
from sqlalchemy import Column, Integer, String, select, ForeignKey
from sqlalchemy.dialects.sqlite import insert
from sqlalchemy.orm import declarative_base, relationship

# Models for each tables
Base = declarative_base()


class SensorData(Base):
    __tablename__ = "sensor_data"

    id = Column(Integer, primary_key=True)
    name = Column(String(200), unique=True)
    value = Column(String)


class Condition(Base):
    __tablename__ = "condition"

    id = Column(Integer, primary_key=True)
    name = Column(String(200))
    description = Column(String)
    parameters = relationship("Parameter", cascade="all, delete")
    actions = relationship("Action", cascade="all, delete")


class Parameter(Base):
    __tablename__ = "parameter"

    id = Column(Integer, primary_key=True)
    order = Column(Integer)
    name = Column(String(200))
    value = Column(String)
    operator = Column(Integer)
    logic = Column(Integer)
    condition_id = Column(Integer, ForeignKey("condition.id"))


class Action(Base):
    __tablename__ = "action"

    id = Column(Integer, primary_key=True)
    value = Column(String)
    condition_id = Column(Integer, ForeignKey("condition.id"))


def mapped_for_upsert(data):
    measurement = data["measurement"]
    fields = data["fields"]
    sensor_datas = {}
    for field, value in fields.items():
        sensor_data = {
            "name": f"{measurement}#{field}",
            "value": value
        }
        sensor_datas[f"{measurement}#{field}"] = sensor_data

    return sensor_datas


def generate_add_sensor_data_sql(data):
    sensor_datas = []
    for sensor in data:
        stmt = insert(SensorData).values(
            name=sensor["name"],
            value=sensor["value"]
        )
        do_update_stmt = stmt.on_conflict_do_update(
            index_elements=['name'],
            set_=dict(value=stmt.excluded.value)
        )
        sensor_datas.append(do_update_stmt)

    return sensor_datas


def generate_get_sensor_data_sql(measurement, field):
    stmt = select(SensorData).where(SensorData.name == f'{measurement}#{field}')

    return stmt



from paho.mqtt import client as paho

from loguru import logger


def initialize(mqtt_service_queue, relay_host, relay_port, relay_username, relay_password, topics):
    try:
        def on_message(cli: paho.Client, userdata, msg):
            mqtt_service_queue.put({
                "name": "publish",
                "topic": msg.topic,
                "payload": msg.payload
            })

        def on_connect(cli: paho.Client, userdata, flags, rc):
            for topic in topics:
                cli.subscribe(topic)

        client = paho.Client()

        client.reconnect_delay_set(max_delay=120)
        client.on_message = on_message
        client.on_connect = on_connect
        client.username_pw_set(relay_username, relay_password)
        client.connect(relay_host, relay_port, 60)

        # Start MQTT main loop
        client.loop_forever()
    except Exception as e:
        logger.exception(f"There is something wrong when executing relayer operation [{e}]")
    finally:
        return



from influxdb_client.client.influxdb_client_async import InfluxDBClientAsync
from loguru import logger
from ..utils import config


async def write(data, bucket=None):
    try:
        if bucket is None:
            bucket = config.homesolar_config['INFLUXDB']['default_bucket']

        async with InfluxDBClientAsync(
                url=f"{config.homesolar_config['INFLUXDB']['host']}:{config.homesolar_config['INFLUXDB']['port']}",
                token=config.homesolar_config['INFLUXDB']['token'],
                org=config.homesolar_config['INFLUXDB']['org']) as client:
            await client.ping()

            write_api = client.write_api()
            success = await write_api.write(bucket=bucket, record=data)
            if success:
                logger.debug(f"Data saved successfully!")
            else:
                logger.error(f"Something went wrong, fail to saved the data!")
    except Exception as e:
        logger.exception(f"Data not saved! [{e}]")


async def query(flux, url=None, token=None, org=None, timeout=None):
    if url is None:
        url = f"{config.homesolar_config['INFLUXDB']['host']}:{config.homesolar_config['INFLUXDB']['port']}"
    if token is None:
        token = config.homesolar_config['INFLUXDB']['token']
    if org is None:
        org = config.homesolar_config['INFLUXDB']['org']
    if timeout is None:
        timeout = config.homesolar_config['INFLUXDB']['read_timeout']

    async with InfluxDBClientAsync(
            url=url,
            token=token,
            org=org,
            timeout=timeout
    ) as client:
        await client.ping()

        query_api = client.query_api()
        result = await query_api.query(flux)
        return result



import json
from time import perf_counter

from loguru import logger

# project's imports
from homesolar.services import influxdb
from homesolar.utils import datetime, config, data_formatter
from homesolar.utils import influxdb as influx_utils
from homesolar.utils import sqlite as sqlite_utils
from homesolar.utils.influxdb import serialize
from homesolar.utils.sqlite import SensorData


# Used on All database
async def write_sensor_data(data):
    from homesolar.services import sqlite
    start_time = perf_counter()
    sensor_data = data_formatter.format_sensor_data(data)
    await influxdb.write(sensor_data)
    logger.debug("is it waiting ?")
    mapped_data = sqlite_utils.mapped_for_upsert(sensor_data)
    await sqlite.bulk_upsert_sensors(mapped_data)
    logger.debug(json.dumps(sensor_data))
    logger.debug(f"Time taken for entire sensor_data_write: {perf_counter() - start_time} second(s)")


async def write_sensor_to_influxdb(data):
    await influxdb.write(data_formatter.format_sensor_data(data))


async def write_sensor_to_sqlite(data):
    from homesolar.services import sqlite
    await sqlite.bulk_upsert_sensors(sqlite_utils.mapped_for_upsert(data_formatter.format_sensor_data(data)))


# Used only on InfluxDB
async def get_battery_charge(date):
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['battery_charge_measurement'],
        config.homesolar_config['DATA']['battery_charge_field'],
        datetime.stringify_timestamp(date),
        datetime.get_next_day(date), "DAY"
    )
    result = await influxdb.query(flux)
    logger.debug(result)
    return data_formatter.simplify_serialized_data(influx_utils.serialize(result, "DAY"))


async def get_battery_usage(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['battery_power_measurement'],
        config.homesolar_config['DATA']['battery_power_field'],
        start_time,
        stop_time,
        timescale
    )
    result = await influxdb.query(flux)
    logger.debug(result)
    return data_formatter.simplify_serialized_data(influx_utils.serialize(result, timescale))


async def get_solar_production(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['solar_production_measurement'],
        config.homesolar_config['DATA']['solar_production_field'],
        start_time,
        stop_time,
        timescale
    )
    result = await influxdb.query(flux)
    logger.debug(result)
    return data_formatter.simplify_serialized_data(influx_utils.serialize(result, timescale))


async def get_grid_usage(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['grid_power_measurement'],
        config.homesolar_config['DATA']['grid_power_field'],
        start_time,
        stop_time,
        timescale
    )
    result = await influxdb.query(flux)
    logger.debug(result)
    return data_formatter.simplify_serialized_data(influx_utils.serialize(result, timescale))


async def get_inverter_usage(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['inverter_power_measurement'],
        config.homesolar_config['DATA']['inverter_power_field'],
        start_time,
        stop_time,
        timescale
    )
    result = await influxdb.query(flux)
    logger.debug(result)
    return data_formatter.simplify_serialized_data(influx_utils.serialize(result, timescale))


async def get_home_usage(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_combined_tables_flux(
        [config.homesolar_config['DATA']['grid_power_measurement'],
         config.homesolar_config['DATA']['inverter_power_measurement']],
        [config.homesolar_config['DATA']['grid_power_field'],
         config.homesolar_config['DATA']['inverter_power_field']],
        start_time,
        stop_time,
        timescale
    )
    logger.debug(flux)
    result = await influxdb.query(flux)
    return data_formatter.simplify_serialized_data(influx_utils.serialize(result, timescale))


async def get_chart_data(date, timescale):
    solar_production = await get_solar_production(date, timescale)
    battery_usage = await get_battery_usage(date, timescale)
    grid_usage = await get_grid_usage(date, timescale)
    inverter_usage = await get_inverter_usage(date, timescale)
    home_usage = await get_home_usage(date, timescale)

    if timescale == "DAY":
        battery_charge = await get_battery_charge(date)
    else:
        battery_charge = None

    data = {
        "solar_production": None if solar_production == {} else solar_production,
        "battery_usage": None if battery_usage == {} else battery_usage,
        "grid_usage": None if grid_usage == {} else grid_usage,
        "inverter_usage": None if inverter_usage == {} else inverter_usage,
        "home_usage": None if home_usage == {} else home_usage,
        "battery_charge": None if battery_charge == {} else battery_charge
    }
    return data


async def get_measurements(bucket=None):
    flux = influx_utils.generate_measurements_flux(bucket)
    tables = await influxdb.query(flux)
    measurements = [row.values["_value"] for table in tables for row in table]

    logger.debug(f"Measurements: {measurements}")
    return measurements


async def get_fields(measurement, bucket=None):
    flux = influx_utils.generate_fields_flux(measurement, bucket)
    tables = await influxdb.query(flux)
    fields = [row.values["_value"] for table in tables for row in table]

    logger.debug(f"Fields: {fields}")
    return fields


async def get_configurations():
    configurations = []
    measurements = await get_measurements()
    logger.debug(measurements)
    for measurement in measurements:
        fields = await get_fields(measurement)
        configuration = {"measurement": measurement, "fields": fields}
        configurations.append(configuration)

    return configurations


# Used only on Sqlite
async def get_sensor_data(measurement, field):
    from homesolar.services import sqlite
    sensor_data = await sqlite.query(SensorData, SensorData.name.in_([f"{measurement}#{field}"]))
    if sensor_data is None:
        logger.warning("No sensor data with specified name found")
        return None
    else:
        for sensor in sensor_data:
            return data_formatter.parse_to_float_if_possible(sensor.value)


class InfluxInterface:
    @staticmethod
    async def get_chart_data(date, timescale):
        grid_power_measurement = config.homesolar_config['DATA']['grid_power_measurement']
        grid_power_field = config.homesolar_config['DATA']['grid_power_field']
        inverter_power_measurement = config.homesolar_config['DATA']['inverter_power_measurement']
        inverter_power_field = config.homesolar_config['DATA']['inverter_power_field']
        solar_production_measurement = config.homesolar_config['DATA']['solar_production_measurement']
        solar_production_field = config.homesolar_config['DATA']['solar_production_field']
        battery_power_measurement = config.homesolar_config['DATA']['battery_power_measurement']
        battery_power_field = config.homesolar_config['DATA']['battery_power_field']
        battery_charge_measurement = config.homesolar_config['DATA']['battery_charge_measurement']
        battery_charge_field = config.homesolar_config['DATA']['battery_charge_field']

        data = {
            "measurements": [
                {
                    "measurement": grid_power_measurement,
                    "field": grid_power_field
                },
                {
                    "measurement": inverter_power_measurement,
                    "field": inverter_power_field
                },
                {
                    "measurement": solar_production_measurement,
                    "field": solar_production_field
                },
                {
                    "measurement": battery_power_measurement,
                    "field": battery_power_field
                },
            ],
            "date": date,
            "timescale": timescale
        }

        multiplier = 1
        if timescale == "DAY":
            data['measurements'].append(
                {
                    "measurement": battery_charge_measurement,
                    "field": battery_charge_field
                }
            )
            multiplier /= 12

        flux = InfluxInterface.multiple_flux_generation(data)

        result = await influxdb.query(flux)
        serialized_result = serialize(result, timescale)

        solar_production = data_formatter.simplify_serialized_data(serialized_result, solar_production_measurement,
                                                                   solar_production_field, multiplier)
        battery_usage = data_formatter.simplify_serialized_data(serialized_result, battery_power_measurement,
                                                                battery_power_field, multiplier)
        grid_usage = data_formatter.simplify_serialized_data(serialized_result, grid_power_measurement,
                                                             grid_power_field, multiplier)
        inverter_usage = data_formatter.simplify_serialized_data(serialized_result, inverter_power_measurement,
                                                                 inverter_power_field, multiplier)
        battery_charge = data_formatter.simplify_serialized_data(serialized_result, battery_charge_measurement,
                                                                 battery_charge_field, multiplier)
        home_usage = data_formatter.combine_simplified_data(grid_usage, inverter_usage)

        chart_data = {
            "solar_production": None if solar_production == {} else solar_production,
            "battery_usage": None if battery_usage == {} else battery_usage,
            "grid_usage": None if grid_usage == {} else grid_usage,
            "inverter_usage": None if inverter_usage == {} else inverter_usage,
            "home_usage": None if home_usage == {} else home_usage,
            "battery_charge": None if battery_charge == {} else battery_charge
        }
        return chart_data

    @staticmethod
    def multiple_flux_generation(data, bucket=None, timezone=None):
        result = ""

        start_time, stop_time = datetime.get_date_pair(data["date"], data["timescale"])
        for index, measurement in enumerate(data["measurements"]):
            flux = InfluxInterface.generate_flux(
                measurement["measurement"],
                measurement["field"],
                start_time,
                stop_time,
                data["timescale"],
                bucket,
                timezone,
                index
            )

            result += flux

        logger.debug(result)
        return result

    @staticmethod
    def generate_flux(measurement, field, start_time, end_time, timescale, bucket=None, timezone=None, number=0):
        if bucket is None:
            bucket = config.homesolar_config['INFLUXDB']['default_bucket']

        if timezone is None:
            timezone = config.homesolar_config['INFLUXDB']['timezone']

        if timescale == "DAY":
            InfluxInterface.generate_flux_day(measurement, field, start_time, end_time, bucket, number)

        result = ""
        if number == 0:
            result += f'''
            import "timezone"
            option location = timezone.location(name: "{timezone}")
            '''

        if timescale == "DAY":
            result += InfluxInterface.generate_flux_day(measurement, field, start_time, end_time, bucket, number)
        elif timescale == "MONTH":
            result += InfluxInterface.generate_flux_day(measurement, field, start_time, end_time, bucket, number)
        elif timescale == "YEAR":
            result += InfluxInterface.generate_flux_day(measurement, field, start_time, end_time, bucket, number)

        logger.debug("Generated flux:", result)
        return result

    @staticmethod
    def generate_flux_day(measurement, field, start_time, end_time, bucket, number=None):
        if number is None:
            number = ""

        result = f"""
            data{number} = from(bucket: "{bucket}")
            |> range(start: {start_time}, stop: {end_time})
            |> filter(fn: (r) => r["_measurement"] == "{measurement}")
            |> filter(fn: (r) => r["_field"] == "{field}")
            |> aggregateWindow(every: 5m, fn: mean, createEmpty: true)
            |> yield(name: "{measurement}_{field}")
        """

        return result

    @staticmethod
    def generate_flux_month(measurement, field, start_time, end_time, bucket, number=None):
        if number is None:
            number = ""

        result = f"""
            data{number} = from(bucket: "{bucket}")
            |> range(start: {start_time}, stop: {end_time})
            |> filter(fn: (r) => r["_measurement"] == "{measurement}")
            |> filter(fn: (r) => r["_field"] == "{field}")
            |> aggregateWindow(every: 1h, fn: mean, createEmpty: false)
            |> aggregateWindow(every: 1d, fn: sum, createEmpty: true)
            |> yield(name: "{measurement}_{field}")
        """

        return result

    @staticmethod
    def generate_flux_year(measurement, field, start_time, end_time, bucket, number=None):
        if number is None:
            number = ""

        result = f"""
            data{number} = from(bucket: "{bucket}")
            |> range(start: {start_time}, stop: {end_time})
            |> filter(fn: (r) => r["_measurement"] == "{measurement}")
            |> filter(fn: (r) => r["_field"] == "{field}")
            |> aggregateWindow(every: 1h, fn: mean, createEmpty: false)
            |> aggregateWindow(every: 1mo, fn: sum, createEmpty: true)
            |> yield(name: "{measurement}_{field}")
        """

        return result



from time import perf_counter

from loguru import logger
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session

from ..utils import sqlite, config
from ..utils.sqlite import SensorData

# Not Async
sqlite_engine = create_engine(f"sqlite://{config.homesolar_config['SQLITE']['database']}", echo=False, connect_args={"check_same_thread": False})
sqlite_session = scoped_session(sessionmaker(bind=sqlite_engine, expire_on_commit=False))


# Async
# sqlite_engine = create_async_engine(f"sqlite+aiosqlite://{config.homesolar_config['SQLITE']['database']}", echo=False)
# sqlite_session = scoped_session(sessionmaker(bind=sqlite_engine, expire_on_commit=False, class_=AsyncSession))


def reinitialize_tables():
    sqlite.Base.metadata.drop_all(sqlite_engine)
    sqlite.Base.metadata.create_all(sqlite_engine)


async def async_reinitialize_tables():
    async with sqlite_engine.begin() as conn:
        await conn.run_sync(sqlite.Base.metadata.drop_all)
        await conn.run_sync(sqlite.Base.metadata.create_all)


async def bulk_upsert_sensors(sensors_data):
    try:
        start_time = perf_counter()
        entries_to_update = []
        entries_to_put = []
        with sqlite_session() as session:

            # Find all customers that needs to be updated and build mappings
            for each in (
                    session.query(SensorData).filter(SensorData.name.in_(sensors_data.keys()))
            ):
                sensor = sensors_data.pop(each.name)
                entries_to_update.append({"id": each.id, "name": sensor["name"], "value": sensor["value"]})

            # Bulk mappings for everything that needs to be inserted
            for sensor in sensors_data.values():
                entries_to_put.append({"name": sensor["name"], "value": sensor["value"]})

            if entries_to_put:
                session.bulk_insert_mappings(SensorData, entries_to_put)
            if entries_to_update:
                session.bulk_update_mappings(SensorData, entries_to_update)
            session.commit()

        # if entries_to_put:
        #     write(generate_add_sensor_data_sql(entries_to_put))
        logger.debug(f"Time taken: {perf_counter() - start_time} second(s)")

    except Exception as e:
        logger.exception(f"SQLITE ERROR [{e}]]")


def write(data):
    try:
        start_time = perf_counter()
        with sqlite_session() as session:
            with session.begin():

                if type(data) is list:
                    for stmt in data:
                        session.execute(stmt)
                else:
                    session.execute(data)
            session.commit()

        logger.debug(f"Time taken: {perf_counter() - start_time} second(s)")
        logger.debug("Data saved successfully!")
    except Exception as e:
        logger.exception(f"Data not saved! [{e}]")


async def query(table, filters):
    result = None
    try:
        with sqlite_session() as session:
            result = session.query(table).filter(filters)
            session.commit()
    except Exception as e:
        logger.exception(f"Failed to query data [{e}]")
    finally:
        return result


async def async_write(data):
    try:
        start_time = perf_counter()
        async with sqlite_session() as session:
            async with session.begin():

                if type(data) is list:
                    for stmt in data:
                        await session.execute(stmt)
                else:
                    await session.execute(data)
            await session.commit()
            await session.close()

        logger.debug(f"Time taken: {perf_counter() - start_time} second(s)")
        logger.debug("Data saved successfully!")
    except Exception as e:
        logger.exception(f"Data not saved! [{e}]")


async def execute(statement):
    result = None
    try:
        with sqlite_session() as session:
            with session.begin():
                result = session.execute(statement)
            session.commit()
    except Exception as e:
        logger.exception(f"Something went wrong when executing an sql! [{e}]")
    finally:
        return result


async def async_execute(statement):
    result = None
    try:
        async with sqlite_session() as session:
            async with session.begin():
                result = await session.execute(statement)
            await session.commit()
            await session.close()
    except Exception as e:
        logger.exception(f"Something went wrong when executing an sql! [{e}]")
    finally:
        return result



import json
from collections.abc import MutableMapping

from loguru import logger


def parse_to_float_if_possible(string):
    result = string
    try:
        result = float(string)
    except Exception as e:
        logger.warning(f"Cannot parse to float [{e}]")
        result = parse_to_bool_if_possible(string)
    finally:
        return result


def parse_to_bool_if_possible(string):
    if string in ["true", "True"]:
        result = True
    else:
        result = string
        logger.warning("Cannot parse to boolean")
    return result


def flatten_dict(d, parent_key='', sep='_'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def format_sensor_data(data):
    try:
        payload = json.loads(data["payload"])
        fields = flatten_dict(payload)
    except json.JSONDecodeError:
        fields = {"value": str(data["payload"].decode("utf-8"))}
    data = {
        "measurement": data["name"],
        "fields": fields
    }
    try:
        data["time"] = int(data["time"])
    except:
        pass

    return data


def simplify_serialized_data(data, measurement=None, field=None, multiplexer=1):
    try:
        simplified_data = {}
        sum_value = 0
        for table in data:
            if measurement is None:
                if table['name'] == "data":
                    simplified_data["name"] = f"{table['measurement']}#{table['field']}"
                    simplified_data["values"] = table["values"]

                    for value in table["values"]:
                        simplified_data["sum"] = None_sum(sum_value, value["value"])
            else:
                if table['measurement'] == measurement and table['field'] == field:
                    if table['name'] == "data":
                        simplified_data["name"] = f"{table['measurement']}#{table['field']}"
                        simplified_data["values"] = table["values"]

                        for value in table["values"]:
                            simplified_data["sum"] = None_sum(sum_value, value["value"])
        if sum_value == 0:
            return {}

        simplified_data["sum"] = simplified_data["sum"] * multiplexer

        return simplified_data
    except Exception as e:
        logger.exception(f"Failed to simplify the data, returning [{e}]")
        return data


def None_sum(*args):
    args = [a for a in args if a is not None]
    return sum(args) if args else None


def combine_simplified_data(data_a, data_b):
    combined_data = {}
    try:
        if data_a is None or data_a == {}:
            if data_b is not None:
                return data_b

        if data_b is None or data_b == {}:
            return data_a

        new_values = []
        for idx, value in enumerate(data_a["values"]):
            new_value = value
            new_value["value"] = None_sum(value["value"], data_b["values"][idx]["value"])
            new_values.append(new_value)

        new_sum = None_sum(data_a["sum"] + data_b["sum"])
        new_name = data_a["name"] + data_b["name"]
        combined_data = {
            "name": new_name,
            "values": new_values,
            "sum": new_sum
        }
    except Exception as e:
        logger.exception(f"Failed to combine data, [{e}]")
    finally:
        return combined_data



import re

from loguru import logger

from ..utils import config


def generate_combined_tables_flux(measurements, fields, start_time, end_time, timescale,
                                  combined_measurement="combined_measurement", combined_field="combined_field",
                                  operator="+", bucket=None):
    if bucket is None:
        bucket = config.homesolar_config['INFLUXDB']['default_bucket']

    additional_query = ""
    yield_str = ""
    aggregate_str = '''
    |> aggregateWindow (every:1m, fn: mean, createEmpty: false)
    |> toFloat()
    |> aggregateWindow(every: 1h, fn: sum, createEmpty: false)
    |> map(fn: (r) => ({r with _value: r._value/60.0}))
    '''

    raw_data_query = f'''
    import "timezone"
    option location = timezone.location(name: "{config.homesolar_config['INFLUXDB']['timezone']}")
    '''
    data_query = ""
    tables = "{"
    values = "{r with _value:"
    regex_filter = re.compile(r'[^a-zA-Z]+')
    add_measurement_field = "{" + f'r with _measurement: "{combined_measurement}", _field:"{combined_field}"' + "}"

    if timescale == "DAY":
        yield_str = '|> yield(name: "formatted_data")'
        aggregate_str = '''
        |> aggregateWindow (every:10s, fn: mean, createEmpty: false)
        |> toFloat()
        |> aggregateWindow(every: 5m, fn: sum, createEmpty: true)
        |> map(fn: (r) => ({r with _value: r._value/30.0}))
        '''
        data_query = '''
        |> aggregateWindow(every: 1h, fn: sum, createEmpty: false)
        |> map(fn: (r) => ({r with _value: r._value/12.0}))
        '''
    elif timescale == "MONTH":
        additional_query = f'''
            formatted_data = data
            |> aggregateWindow(every: 1d, fn: sum, createEmpty: true)
            |> map(fn: (r) => ({add_measurement_field}))
            |> yield(name: "formatted_data")
        '''
    elif timescale == "YEAR":
        additional_query = f'''
            formatted_data = data
            |> aggregateWindow(every: 1mo, fn: sum, createEmpty: true)
            |> map(fn: (r) => ({add_measurement_field}))
            |> yield(name: "formatted_data")
        '''

    for measurement, field in zip(measurements, fields):
        name = re.sub(regex_filter, "", f'{measurement}_{field}')
        raw_data_query += f'''
        {name} = from(bucket: "{bucket}")
        |> range(start: {start_time}, stop: {end_time})
        |> filter(fn: (r) => r["_measurement"] == "{measurement}")
        |> filter(fn: (r) => r["_field"] == "{field}")
        {aggregate_str}
        '''

        tables += f"{name}:{name},"
        values += f"r._value_{name}{operator}"

    tables = tables[:-1] + "}"
    values = values[:-1]

    result = f'''
    {raw_data_query}

    combined_value = join(tables: {tables}, on:["_time"])
    |> map(fn:(r) => ({values},_measurement:"{combined_measurement}",_field:"{combined_field}"{'}'}))
    {yield_str}

    data = combined_value
    {data_query}

    {additional_query}

    sum = data
    |> sum()
    |> yield(name: "sum")
    '''

    logger.debug(result)
    return result


def generate_flux(measurement, field, start_time, end_time, timescale, bucket=None):
    if bucket is None:
        bucket = config.homesolar_config['INFLUXDB']['default_bucket']

    data_query = ""
    additional_query = ""
    yield_str = '''
    |> aggregateWindow (every:1m, fn: mean, createEmpty: false)
    |> toFloat()
    |> aggregateWindow(every: 1h, fn: sum, createEmpty: false)
    |> map(fn: (r) => ({r with _value: r._value/60.0}))
    '''

    if timescale == "DAY":
        yield_str = '''
        |> aggregateWindow (every:10s, fn: mean, createEmpty: false)
        |> toFloat()
        |> aggregateWindow(every: 5m, fn: sum, createEmpty: true)
        |> map(fn: (r) => ({r with _value: r._value/30.0}))
        |> yield(name: "formatted_data")
        '''
        data_query = '''
        |> aggregateWindow(every: 1h, fn: sum, createEmpty: false)
        |> map(fn: (r) => ({r with _value: r._value/12.0}))
        '''
    elif timescale == "MONTH":
        additional_query = f'''
        formatted_data = data
        |> aggregateWindow(every: 1d, fn: sum, createEmpty: true)
        |> yield(name: "formatted_data")
        '''
    elif timescale == "YEAR":
        additional_query = f'''
        formatted_data = data
        |> aggregateWindow(every: 1mo, fn: sum, createEmpty: true)
        |> yield(name: "formatted_data")
        '''

    result = f'''
    import "timezone"
    option location = timezone.location(name: "{config.homesolar_config['INFLUXDB']['timezone']}")
        
    raw_data = from(bucket: "{bucket}")
    |> range(start: {start_time}, stop: {end_time})
    |> filter(fn: (r) => r["_measurement"] == "{measurement}")
    |> filter(fn: (r) => r["_field"] == "{field}")
    {yield_str}

    data = raw_data
    {data_query}

    {additional_query}

    sum = data
    |> sum()
    |> yield(name: "sum")
    '''
    logger.debug(result)
    return result


def generate_measurements_flux(bucket=None):
    if bucket is None:
        bucket = config.homesolar_config['INFLUXDB']['default_bucket']

    query = f'''
    import "influxdata/influxdb/schema"
    schema.measurements(bucket: "{bucket}")
    '''

    return query


def generate_fields_flux(measurement, bucket=None):
    if bucket is None:
        bucket = config.homesolar_config['INFLUXDB']['default_bucket']

    query = f'''
        import "influxdata/influxdb/schema"
        schema.measurementFieldKeys(bucket: "{bucket}", measurement: "{measurement}")
        '''

    return query


def serialize(tables, timescale="DAY"):
    data = []
    try:
        for idx, table in enumerate(tables):
            field = {"table": idx}
            values = []
            for row in table:
                value = {}
                try:
                    field["measurement"] = row.values["_measurement"]
                    field["field"] = row.values["_field"]
                    value["time"] = row.values["_time"]
                except Exception as e:
                    logger.warning(f"Something went wrong when serializing data, skipping... [{e}]")

                value["value"] = row.values["_value"]
                values.append(value)

            # Check length of values saved
            field["name"] = check_table_name(len(values))
            field["values"] = values

            data.append(field)
    except Exception as e:
        logger.error(e)
    finally:
        return data


def check_table_name(length: int):
    if length == 1:
        return "sum"

    return "data"



import json
from ..utils.mqtt import MqttTopic
from ..interfaces import database
from ..utils import datetime, data_formatter, config, bluetooth

today_production = 0
today_independence = 0


async def send_summary(mqtt_queue, update=False):
    global today_independence, today_production

    payload = {
        "solar_production": await database.get_sensor_data(
            measurement=config.homesolar_config['DATA']['solar_production_measurement'],
            field=config.homesolar_config['DATA']['solar_production_field']
        ), "battery_usage": await database.get_sensor_data(
            measurement=config.homesolar_config['DATA']['battery_power_measurement'],
            field=config.homesolar_config['DATA']['battery_power_field']
        ), "battery_charge": await database.get_sensor_data(
            measurement=config.homesolar_config['DATA']['battery_charge_measurement'],
            field=config.homesolar_config['DATA']['battery_charge_field']
        ), "grid_usage": await database.get_sensor_data(
            measurement=config.homesolar_config['DATA']['grid_power_measurement'],
            field=config.homesolar_config['DATA']['grid_power_field']
        ), "inverter_usage": await database.get_sensor_data(
            measurement=config.homesolar_config['DATA']['inverter_power_measurement'],
            field=config.homesolar_config['DATA']['inverter_power_field']
        )}

    if payload["grid_usage"] is None or payload["inverter_usage"] is None:
        payload["home_usage"] = None
    else:
        payload["home_usage"] = payload["grid_usage"] + payload["inverter_usage"]

    if update:
        payload["today_production"] = \
            today_production = (await database.get_solar_production(date=datetime.get_today(), timescale="DAY"))["sum"]
        grid = (await database.get_grid_usage(date=datetime.get_today(), timescale="DAY"))["sum"]
        inverter = (await database.get_inverter_usage(date=datetime.get_today(), timescale="DAY"))["sum"]
        payload["today_independence"] = today_independence = (inverter / (grid + inverter)) * 100.0
    else:
        payload["today_production"] = today_production
        payload["today_independence"] = today_independence

    task = {
        "name": "publish",
        "topic": MqttTopic.SUMMARY,
        "payload": json.dumps(payload),
        "retain": True
    }
    mqtt_queue.put(task)


async def send_battery(mqtt_queue):
    payload = {
        "status": await database.get_sensor_data(
            measurement=config.homesolar_config['DATA']['battery_status_measurement'],
            field=config.homesolar_config['DATA']['battery_status_field']
        ), "pack_voltage": await database.get_sensor_data(
            measurement=config.homesolar_config['DATA']['battery_voltage_measurement'],
            field=config.homesolar_config['DATA']['battery_voltage_field']
        ), "pack_amperage": await database.get_sensor_data(
            measurement=config.homesolar_config['DATA']['battery_amperage_measurement'],
            field=config.homesolar_config['DATA']['battery_amperage_field']
        ), "pack_power": await database.get_sensor_data(
            measurement=config.homesolar_config['DATA']['battery_power_measurement'],
            field=config.homesolar_config['DATA']['battery_power_field']
        ), "soc": await database.get_sensor_data(
            measurement=config.homesolar_config['DATA']['battery_charge_measurement'],
            field=config.homesolar_config['DATA']['battery_charge_field']
        )}

    if payload["status"] is not None:
        payload["status"] = bluetooth.MOSFET_Discharge_St[int(payload["status"])]

    single_cell = config.homesolar_config['DATA']['battery_single_cell']
    formatted_cells = []

    cells = config.homesolar_config['DATA']['battery_cells_fields'].split()
    balances = config.homesolar_config['DATA']['battery_cells_balance_fields'].split()
    if not single_cell:
        for index, cell in enumerate(cells):
            formatted_cell = {
                "name": f"Cell {index+1}",
                "balance": await database.get_sensor_data(
                    measurement=config.homesolar_config['DATA']['battery_cells_balance_measurement'],
                    field=balances[index]
                ) == 1,
                "voltage": await database.get_sensor_data(
                    measurement=config.homesolar_config['DATA']['battery_cells_measurement'],
                    field=cells[index]
                )
            }
            formatted_cells.append(formatted_cell)

    payload["cells"] = formatted_cells

    task = {
        "name": "publish",
        "topic": MqttTopic.BATTERY,
        "payload": json.dumps(payload),
        "retain": True
    }
    mqtt_queue.put(task)




import json
import time

from loguru import logger

from homesolar.utils.sqlite import Condition, SensorData, Parameter, Action


def control_loop(timeout):
    logger.debug("Controls Started")
    create_dummy_data()
    try:
        elapsed_time = 0
        while True:
            if time.perf_counter() - elapsed_time >= timeout:
                elapsed_time = time.perf_counter()
                loop()
                logger.debug("Controls Checked")
    except Exception as e:
        logger.exception(f"Something went wrong when looping through controls [{e}]")


def check_controls():
    conditions = get_conditions()
    for condition in conditions:
        if check_parameters(condition.parameters):
            for action in condition.actions:
                do_action(action)


def check_parameters(parameters):
    results = []
    for parameter in parameters:
        value1 = parse_to_float(parameter.value)
        value2 = parse_to_float(get_sensor_data(parameter.name))
        results.append(compare_values(parameter.operator, value1, value2))

    result = results[0]
    for idx, parameter in enumerate(parameters):
        if idx == len(results):
            break
        # AND
        if parameter.logic == 1:
            result = result and results[idx + 1]
        # OR
        if parameter.logic == 2:
            result = result or results[idx + 1]

    return result


def do_action(action):
    value = json.loads(action.value)
    if value["type"] == "MQTT":
        task = {
            "name": "publish",
            "topic": value["topic"],
            "payload": value["payload"]
        }
        logger.debug(task)
        return
    if value["type"] == "GPIO":
        task = {
            "name": "gpio",
            "value": value["payload"]
        }
        logger.debug(task)
        return
    return


def get_conditions():
    return []


def get_sensor_data(sensor_name):
    return 0


def parse_to_float(value):
    try:
        return float(value)
    except:
        return value


def compare_values(operator, value1, value2):
    # Equals
    if operator == 1:
        if value1 == value2:
            return True
    # Not Equals
    if operator == 2:
        if value1 != value2:
            return True
    # Greater
    if operator == 3:
        if value1 > value2:
            return True
    # Less
    if operator == 4:
        if value1 < value2:
            return True
    # Greater or Equals
    if operator == 5:
        if value1 >= value2:
            return True
    # Less or Equals
    if operator == 6:
        if value1 <= value2:
            return True

    return False


def create_dummy_data():
    from homesolar.services.sqlite import sqlite_session
    with sqlite_session() as session:
        param_a = Parameter(name="TasmotaSolar#Power", operator=2, value="2000asdasd00", logic=1)
        param_b = Parameter(name="TasmotaSolar#Power", operator=2, value="1000", logic=2)
        param_c = Parameter(name="BMS#Power", operator=2, value="1000", logic=1)
        param_d = Parameter(name="BMS#Power", operator=2, value="2000", logic=2)
        param_e = Parameter(name="BMS#Power", operator=2, value="2000", logic=-1)

        action_a = Action(value="Turn On")
        action_b = Action(value="Turn OFF")
        action_c = Action(value="Toggle")

        session.add_all([param_a, param_b, param_c, param_d, param_e, action_a, action_b, action_c])
        session.commit()

        condition_a = Condition(
            name="ConditionA",
            description="Condition A Description",
            parameters=[
                param_a, param_b, param_c, param_d, param_e
            ],
            actions=[
                action_a, action_b, action_c
            ]
        )

        session.add_all([condition_a])
        session.commit()


def loop():
    from homesolar.services.sqlite import sqlite_session
    while True:
        with sqlite_session() as session:
            conditions = session.query(Condition)
            logger.debug(conditions)
            for condition in conditions:
                condition_resolve = ""
                for parameter in condition.parameters:
                    data = session.query(SensorData).filter(SensorData.name.in_([str(parameter.name)]))
                    if data:
                        condition_resolve += str(
                            compare(try_parse_to_float(data.value), try_parse_to_float(parameter.value),
                                    parameter.operator))
                    else:
                        condition_resolve += "False"
                    if parameter.logic != -1:
                        condition_resolve += f" {parameter.logic} "
                    else:
                        break

            break


def resolve_condition(resolve: str) -> bool:
    items = resolve.split(" ")

    result = False
    operator = -1
    for index, item in enumerate(items):
        parsed_item = parse_resolve_item(item)

        if index == 0:
            result = parsed_item
        elif parsed_item == 1:
            operator = parsed_item
        elif parsed_item == 2:
            operator = parsed_item
        elif operator != "":
            if operator == 1:
                result = result and parsed_item
            else:
                result = result or parsed_item

    logger.debug(f"({resolve}) resolved as", result)
    return result


def parse_resolve_item(item: str) -> any:
    if item == "True":
        return True
    elif item == "False":
        return False
    return item


def try_parse_to_float(value: any) -> any:
    try:
        return float(value)
    except ValueError:
        return value


def compare(v1: any, v2: any, operator: int) -> bool:
    if not isinstance(v1, type(v2)):
        return False

    if operator == 1:
        if v1 > v2:
            return True
    elif operator == 2:
        if v1 >= v2:
            return True
    elif operator == 3:
        if v1 < v2:
            return True
    elif operator == 4:
        if v1 <= v2:
            return True
    elif operator == 5:
        if v1 == v2:
            return True
    elif operator == 6:
        if v1 != v2:
            return True
    return False



from datetime import datetime, timedelta, date
from json import JSONEncoder
import pytz


class DateTimeEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime, date)):
            return obj.timestamp()


def get_timezone(timezone=None):
    from ..utils import config
    if timezone is None:
        timezone = config.homesolar_config['INFLUXDB']['timezone']
    return pytz.timezone(timezone)


def stringify_timestamp(timestamp):
    return datetime.utcfromtimestamp(timestamp).strftime('%Y-%m-%dT%H:%M:%SZ')


def get_today(timezone=None):
    return datetime.now(get_timezone(timezone)).replace(hour=0, minute=0, second=0).timestamp()


def get_next_day(timestamp):
    return (datetime.utcfromtimestamp(timestamp) + timedelta(days=1)).strftime('%Y-%m-%dT%H:%M:%SZ')


def get_first_day_of_month(timestamp,timezone=None):
    return datetime.fromtimestamp(timestamp, get_timezone(timezone)) \
        .replace(day=1) \
        .astimezone(pytz.utc) \
        .strftime('%Y-%m-%dT%H:%M:%SZ')


def get_last_day_of_month(timestamp, timezone=None):
    next_month = datetime.fromtimestamp(timestamp, get_timezone(timezone)).replace(day=28) + timedelta(days=4)
    return (next_month - timedelta(days=next_month.day) + timedelta(days=1)) \
        .astimezone(pytz.utc) \
        .strftime('%Y-%m-%dT%H:%M:%SZ')


def get_first_day_of_year(timestamp, timezone=None):
    return datetime.fromtimestamp(timestamp, get_timezone(timezone)) \
        .replace(day=1, month=1) \
        .astimezone(pytz.utc) \
        .strftime('%Y-%m-%dT%H:%M:%SZ')


def get_last_day_of_year(timestamp, timezone=None):
    return (datetime.fromtimestamp(timestamp, get_timezone(timezone))
            .replace(day=31, month=12) + timedelta(days=1)) \
        .astimezone(pytz.utc) \
        .strftime('%Y-%m-%dT%H:%M:%SZ')


def get_date_pair(date, timescale, timezone=None):
    if timescale == "MONTH":
        start_time = get_first_day_of_month(date, timezone)
        stop_time = get_last_day_of_month(date, timezone)
    elif timescale == "YEAR":
        start_time = get_first_day_of_year(date, timezone)
        stop_time = get_last_day_of_year(date, timezone)
    else:
        start_time = stringify_timestamp(date)
        stop_time = get_next_day(date)

    return start_time, stop_time
